####################################################################################################
# VERSION HISTORY                                                                                  #
# ===============                                                                                  #
#                                                                                                  #
# Version   | Author           | Description of changes                      |  Reviewed by / Date #
# ----------+------------------|---------------------------------------------+---------------------#
# 2010-11-26|Philippe PFISTER  | Initial release                             | /                   #
# ----------+------------------+---------------------------------------------+---------------------#
#                                                                                                  #
####################################################################################################
#  PURPOSE                                                                                         #
#  =======                                                                                         #
#  Driver for an LCD display based on the M50530                                                   #
#  Synopsis: cf. Test_SII_C591003_Display sub.                                                     #                                         #
#  Test / debug command line user interface: cf. TestInterface sub.                                #
####################################################################################################
#  COPYRIGHT                                                                                       #
#  =========                                                                                       #
# Copyright 2011 Philippe Pfister                                                                  #
#                                                                                                  #
# Licensed under the EUPL (Free/Open Source Software Licence), Version 1.1 only (the "Licence");   #
# You may not use this work except in compliance with the Licence.                                 #
# You may obtain a copy of the Licence at:                                                         #
#                                                                                                  #
# http://www.osor.eu/eupl/european-union-public-licence-eupl-v.1.1                                 #                                                               #
#                                                                                                  #
# Unless required by applicable law or agreed to in writing, software distributed under the        #
# Licence is distributed on an "AS IS" basis,                                                      #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                         #
# See the Licence for the specific language governing permissions and limitations under the        #
# Licence.                                                                                         #
####################################################################################################
# NOTES                                                                                            #
# =====                                                                                            #
# This program needs the following modules (not included):                                         #
# - Device::ParallelPort::drv::win32                                                               #
#   (on a windows32 system, alternatives are: Device::ParallelPort::drv::(linux|parport|win32))    #
# - Device::ParallelPort                                                                           #
# To install those, type CPAN at the DOS prompt, then type:                                        #
# install 'Device::ParallelPort'                                                                   #
# install 'Device::ParallelPort::drv::win32'                                                       #
# q                                                                                                #
#                                                                                                  #
# Furthermore, you will need inpout32.dll (not included, free non-commercial use) from             #
# http://www.logix4u.net/inpout32.htm                                                              #
####################################################################################################

use strict;
use Time::HiRes qw( gettimeofday tv_interval );
use Device::ParallelPort::drv::win32;
use Device::ParallelPort::drv::auto;
use Device::ParallelPort;

# --- Constants ---

# M50530 Instruction set (instruction code bit 0 = R/W, bit 1 = I/OC2, bit 2 = I/OC1)
*MODE = \0;
*READ_FLAGS = \1;
*WRITE_RAM_DATA = \2;
*READ_RAM_DATA = \3;
*WRITE_DISPLAY_START_ADDR = \4;
*READ_DISPLAY_START_ADDR = \5;
*WRITE_CURSOR_ADDR = \6;
*READ_CURSOR_ADDR = \7;

# Mode constants (DB0..DB7)
*SET_FUNCTION_MODE = \0xC0;
*SET_ENTRY_MODE = \0x40;
*SET_DISPLAY_MODE = \0x20;
*SHIFT_START_ADDR = \0x10;
*SET_UNDERLINE_MODE = \0x0C;
*WRITE_UNDERLINE_BIT = \0x08;
*SET_BLINK_FREQ = \0x04;
*HOME = \0x03;
*CLEAR_HOME = \0x01;
*NOP = \0x00;

*FLAG_BUSY_HI = \0x80;  # Busy flag bit in the higher quartet
*FLAG_BUSY_LO = \0x08;  # Busy flag bit in the lower quartet
*FLAG_4u2_HI = \0x40;  # 4u2 flag bit in the higher quartet
*FLAG_4u2_LO = \0x04;  # 4u2 flag bit in the lower quartet
*FLAG_IO8_4 = \0x20;  # I/O 8/4 flag bit
*FLAG_FONT8_12 = \0x10;  # FONT 8/12 flag bit
*FLAG_RA1 = \0x02;  # RAM region flag 1
*FLAG_RA0 = \0x01;  # RAM region flag 0

# LCD Error codes
*LCD_NO_ERR = \0;  # No error
*LCD_ERR_TIMEOUT = \1;  # The last instruction timed out

# --- Variables ---

# Global adjustable parameters
my $WaitSub = \&WaitDoNothing;  # (pointer to a sub): a pointer to a sub to be regularly called while
  # waiting for the M50530's busy flag to get low
my $WaitSubParms = \0;  # (pointer): a pointer to parameters used for calling the WaitSub
my $TimeOut = 3;  # (float): number of floating seconds to wait for the M50530's busy flag to get
  # low before returning a timeout error code

# Physical port mapping to M50530 signals (pointers to subs)
my $M50530_RW_SET = \&NoInterface;  # Set M50530 R/W pin
my $M50530_RW_GET = \&NoInterface;  # Read M50530 R/W pin
my $M50530_IOC2_SET = \&NoInterface;  # Set M50530 I/OC2 pin
my $M50530_IOC2_GET = \&NoInterface;  # Read M50530 I/OC2 pin
my $M50530_IOC1_SET = \&NoInterface;  # Set M50530 I/OC1 pin
my $M50530_IOC1_GET = \&NoInterface;  # Read M50530 I/OC1 pin
my $M50530_EX_SET = \&NoInterface;  # Set M50530 EX pin
my $M50530_EX_GET = \&NoInterface;  # Read M50530 EX pin
my $M50530_DB0_SET = \&NoInterface;  # Set M50530 DB0 pin
my $M50530_DB0_GET = \&NoInterface;  # Read M50530 DB0 pin
my $M50530_DB1_SET = \&NoInterface;  # Set M50530 DB1 pin
my $M50530_DB1_GET = \&NoInterface;  # Read M50530 DB1 pin
my $M50530_DB2_SET = \&NoInterface;  # Set M50530 DB2 pin
my $M50530_DB2_GET = \&NoInterface;  # Read M50530 DB2 pin
my $M50530_DB3_SET = \&NoInterface;  # Set M50530 DB3 pin
my $M50530_DB3_GET = \&NoInterface;  # Read M50530 DB3 pin
my $M50530_DB4_SET = \&NoInterface;  # Set M50530 DB4 pin
my $M50530_DB4_GET = \&NoInterface;  # Read M50530 DB4 pin
my $M50530_DB5_SET = \&NoInterface;  # Set M50530 DB5 pin
my $M50530_DB5_GET = \&NoInterface;  # Read M50530 DB5 pin
my $M50530_DB6_SET = \&NoInterface;  # Set M50530 DB6 pin
my $M50530_DB6_GET = \&NoInterface;  # Read M50530 DB6 pin
my $M50530_DB7_SET = \&NoInterface;  # Set M50530 DB7 pin
my $M50530_DB7_GET = \&NoInterface;  # Read M50530 DB7 pin

my $ParPort;  # Parallel port device

sub ParallelPortInterfaceInit
# Parallel port physical 4-bit mode interface to M50530 driver initialization sub
# M50530 signals are mapped as follows:
#
# DB25 pin |  DB25 Signal    | M50530 Signal | M50530 Pin
# ---------+-----------------+---------------+-----------
#  08..25  |     Ground      |    VSS (0V)   |     50
#   N.A.   |      N.A.       |    VDD (5V)   |     49
#    02    |       D0        |     I/OC1     |     42
#    03    |       D1        |     I/OC2     |     41
#    04    |       D2        |      EX       |     39
#    05    |       D3        |      R/W      |     40
#    06    |       D4        |      DB4      |     35
#    10    |      ACK        |      DB4      |     35
#    07    |       D5        |      DB5      |     36
#    11    |     BUSY*       |      DB5      |     36
#    08    |       D6        |      DB6      |     37
#    12    |       PE        |      DB6      |     37
#    09    |       D7        |      DB7      |     38
#    13    |      SEL        |      DB7      |     38
#
# *: BUSY is inverted (i.e. reading the bit from the port returns 1 when the pin is at 0v)
#
# NOTE:
# As the M50530 DB4..DB7 pins are used for input and output signals, an appropriate hardware
# interface is needed between the parallel port and the M50530 in order to avoid short-circuits
# between the parallel port's D4..D7 and the M50530's DB4..DB7 signals.
# This driver assumes a 3 state buffer on D0..D7, where D0..D3 are enabled when STROBE is low
# (i.e. when the corresponding bit of the control port is 1, as STROBE is inverted), and D4..D7
# are enabled via SELECT-PRINTER (also inverted)
#
# Parameters:
#   $InitParms (pointer to a hash of parameters): The parameters for the interface initialization
#     (ignored as the default interface init. sub. doesn't take any parameters. This can be set to null)
# Returns:
#   $Result: the return value of the interface initialization routine (always 1)
{
  my ($InitParms) = @_;  # Get the parameters
  # Initialize the device driver
  $ParPort = Device::ParallelPort->new();
  # Set all parallel port output pins to low
  $ParPort->set_control(chr(0));
  $ParPort->set_data(chr(0));
  # Set the default sub for all instructions
  $M50530_RW_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      # TRUE = Read (R/W = VCC)
      $ParPort->set_bit(19, 0);  # Set SELECT-PRINTER to high (inverted) to disable D4..D7
      $ParPort->set_bit(3, 1);  # Set D3 (R/W) to high = Read
    }
    else
    {
      # FALSE = Write (R/W = Gnd)
      $ParPort->set_bit(3, 0);  # Set D3 (R/W) to low = Write
      $ParPort->set_bit(19, 1);  # Set SELECT-PRINTER to low (inverted) to enable D4..D7
    }
  };
  # We don't support M50530_RW_GET
  $M50530_IOC2_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(1, 1);  # Set D1 to high
    }
    else
    {
      $ParPort->set_bit(1, 0);  # Set D1 to low
    }
  };
  # We don't support M50530_IOC2_GET
  $M50530_IOC1_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(0, 1);  # Set D0 to high
    }
    else
    {
      $ParPort->set_bit(0, 0);  # Set D0 to low
    }
  };
  # We don't support M50530_IOC1_GET
  $M50530_EX_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(2, 1);  # Set D2 to high
    }
    else
    {
      $ParPort->set_bit(2, 0);  # Set D2 to low
    }
  };
  # We don't support M50530_EX_GET
  # We don't support M50530_DB0_SET
  # We don't support M50530_DB0_GET
  # We don't support M50530_DB1_SET
  # We don't support M50530_DB1_GET
  # We don't support M50530_DB2_SET
  # We don't support M50530_DB2_GET
  # We don't support M50530_DB3_SET
  # We don't support M50530_DB3_GET
  $M50530_DB4_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(4, 1);  # Set D4 to high
    }
    else
    {
      $ParPort->set_bit(4, 0);  # Set D4 to low
    }
  };
  $M50530_DB4_GET = sub{
    return ((ord($ParPort->get_status) & 0x40) != 0);  # Return the parallel port's ACK signal level
    };
  $M50530_DB5_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(5, 1);  # Set D5 to high
    }
    else
    {
      $ParPort->set_bit(5, 0);  # Set D5 to low
    }
  };
  $M50530_DB5_GET = sub{
    return ((ord($ParPort->get_status) & 0x80) == 0);  # Return the parallel port's BUSY signal (inverted)
    };
  $M50530_DB6_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(6, 1);  # Set D6 to high
    }
    else
    {
      $ParPort->set_bit(6, 0);  # Set D6 to low
    }
  };
  $M50530_DB6_GET = sub{
    return ((ord($ParPort->get_status) & 0x20) != 0);  # Return the parallel port's PE signal level
    };
  $M50530_DB7_SET = sub{
    my ($BitVal) = @_;  # Get the parameter
    if ($BitVal)
    {
      $ParPort->set_bit(7, 1);  # Set D7 to high
    }
    else
    {
      $ParPort->set_bit(7, 0);  # Set D7 to low
    }
  };
  $M50530_DB7_GET = sub{
    return ((ord($ParPort->get_status) & 0x10) != 0);  # Return the parallel port's SEL signal level
    };
  $ParPort->set_bit(16, 1);  # Set STROBE to Gnd (inverted) to enable D0..D3
  return 1;  # Success
}

sub NoInterfaceInit
# Default physical interface driver initialization sub
# Assigns the "NoInterface" sub to all M50530 signals
# Parameters:
#   $InitParms (pointer to a hash of parameters): The parameters for the interface initialization
#     (ignored as the default interface init. sub. doesn't take any parameters. This can be set to null)
# Returns:
#   $Result: the return value of the interface initialization routine (always 1)
{
  my ($InitParms) = @_;  # Get the parameters
  # Set the default sub for all instructions
  $M50530_RW_SET = \&NoInterface;
  $M50530_RW_GET = \&NoInterface;
  $M50530_IOC2_SET = \&NoInterface;
  $M50530_IOC2_GET = \&NoInterface;
  $M50530_IOC1_SET = \&NoInterface;
  $M50530_IOC1_GET = \&NoInterface;
  $M50530_EX_SET = \&NoInterface;
  $M50530_EX_GET = \&NoInterface;
  $M50530_DB0_SET = \&NoInterface;
  $M50530_DB0_GET = \&NoInterface;
  $M50530_DB1_SET = \&NoInterface;
  $M50530_DB1_GET = \&NoInterface;
  $M50530_DB2_SET = \&NoInterface;
  $M50530_DB2_GET = \&NoInterface;
  $M50530_DB3_SET = \&NoInterface;
  $M50530_DB3_GET = \&NoInterface;
  $M50530_DB4_SET = \&NoInterface;
  $M50530_DB4_GET = \&NoInterface;
  $M50530_DB5_SET = \&NoInterface;
  $M50530_DB5_GET = \&NoInterface;
  $M50530_DB6_SET = \&NoInterface;
  $M50530_DB6_GET = \&NoInterface;
  $M50530_DB7_SET = \&NoInterface;
  $M50530_DB7_GET = \&NoInterface;
  return 1;
}

sub NoInterface
# Default interface sub to M50530 signals when no physical interface has been initialized
# Does nothing but returning 0
# Parameters:
#   $Level (boolean): Ignored for *_GET subs. For *_SET subs, 0 sets the M50530 pin to low, any
#     other value sets it to high.
# Returns:
#   $RetData (boolean): the data returned by the M50530 in case of *_GET subs (0 for low, 1 for
#     high), 0 in case of write operations. The NoInterface sub always returns 0.
{
  return 0;
}

sub SetInterface
# Initializes and sets the indicated interface as the current interface to use for M50530 operations
# Parameters:
#   $Interface (pointer to an interface initialisation sub): The interface to activate
#   $InitParms (pointer to a hash of parameters): The parameters for the interface initialization
#     (interface dependent)
# Returns:
#   $Elapsed: the actual number of elapsed nanoseconds
{
  my ($Interface, $InitParms) = @_;  # Get the parameters
  # Set the default sub for all instructions
  NoInterfaceInit(\0);
  # Call the interface init routine and return the result
  return &$Interface($InitParms);
}

sub nsDelay
# Waits at least for the given number of nanoseconds before returning
# Parameters:
#   $ns: The time to wait in nanoseconds
{
  my $StartTime;
  my $S;
  my $I;

  my ($ns) = @_;  # Get the parameters

  $StartTime = [gettimeofday];
  $S = $ns / 1000000000;  # Convert ns to floating seconds
  do
  {
    $I = tv_interval($StartTime);
  } until ($I >= $S);
  return $I * 1000000000;  # Return the actual number of elapsed nanoseconds
}

sub RawExecuteInstr
# Low level access to the M50530
# The instructions are sent regardless of the display mode (4/8-bit) and busy status
# Sends the given instruction and data to the M50530 and returns the M50530 data in case of a read instruction
# Parameters:
#   $Instr: The M50530 instruction to execute (Cf. constants / M50530 Instruction set, e.g. $main::READ_CURSOR_ADDR)
#   $Data: Data to send to the M50530 in case of a write operation, disregarded in case of read operations
# Returns:
#   $RetData: the data returned by the M50530 in case of read operations, $main::LCD_NO_ERR in case of write operations
# Notes:
#   M50530 data is always passed to/from this sub using 8 bits, whether the M50530 is in 4-bit or 8-bit mode
{
  my ($Instr, $Data) = @_;  # Get the parameters
  my $RetData;  # Return value

  &$M50530_RW_SET($Instr & 0x01);  # Set the R/W bit on the M50530
  &$M50530_IOC2_SET($Instr & 0x02);  # Set the I/OC2 bit on the M50530
  &$M50530_IOC1_SET($Instr & 0x04);  # Set the I/OC1 bit on the M50530

  # Test the R/W bit
  if ($Instr & 0x01)
  {
    # This is a read instruction (R/W = 1)
    $RetData = 0;  # Initialise for setting bits
    # Keep the data bus low for reading the M50530
    &$M50530_DB0_SET(0);
    &$M50530_DB1_SET(0);
    &$M50530_DB2_SET(0);
    &$M50530_DB3_SET(0);
    &$M50530_DB4_SET(0);
    &$M50530_DB5_SET(0);
    &$M50530_DB6_SET(0);
    &$M50530_DB7_SET(0);
    # Execute the command on the M50530
    nsDelay(200);  # Control data setup time
    &$M50530_EX_SET(1);
    nsDelay(300);  # Data output delay time
    # Read D0..D7 from the M50530
    $RetData = $RetData | 0x01 if &$M50530_DB0_GET;
    $RetData = $RetData | 0x02 if &$M50530_DB1_GET;
    $RetData = $RetData | 0x04 if &$M50530_DB2_GET;
    $RetData = $RetData | 0x08 if &$M50530_DB3_GET;
    $RetData = $RetData | 0x10 if &$M50530_DB4_GET;
    $RetData = $RetData | 0x20 if &$M50530_DB5_GET;
    $RetData = $RetData | 0x40 if &$M50530_DB6_GET;
    $RetData = $RetData | 0x80 if &$M50530_DB7_GET;
    &$M50530_EX_SET(0);  # Set the EX pin to low
    nsDelay(300);  # Ensure at least 800ns EX signal interval
  }
  else
  {
    # This is a write instruction (R/W = 0)
    $RetData = $main::LCD_NO_ERR;  # Always return $main::LCD_NO_ERR
    # Set D0..D7 on the M50530
    &$M50530_DB0_SET($Data & 0x01);
    &$M50530_DB1_SET($Data & 0x02);
    &$M50530_DB2_SET($Data & 0x04);
    &$M50530_DB3_SET($Data & 0x08);
    &$M50530_DB4_SET($Data & 0x10);
    &$M50530_DB5_SET($Data & 0x20);
    &$M50530_DB6_SET($Data & 0x40);
    &$M50530_DB7_SET($Data & 0x80);
    # Execute the command on the M50530
    nsDelay(200);  # Control data setup time
    &$M50530_EX_SET(1);
    nsDelay(300);  # EX signal pulse width (min. 200ns)
    &$M50530_EX_SET(0);  # Set the EX pin to low
    nsDelay(300);  # Ensure at least 800ns EX signal interval and 100ns data hold time
  }

  return $RetData;
}

sub WaitDoNothing
# Default sub to be called regularly while waiting for the M50530
# Parameters:
#   $Parms (pointer): a pointer to parameters (ignored)
{
  my ($Parms) = @_;  # Get the parameters
  # Do nothing...
}

sub WaitForReady
# Wait for the busy flag of the M50530 to be low
# (Supports both the 4 and 8-bit mode)
# Parameters:
#   None (uses global parameters $WaitSub, $WaitSubParms, $TimeOut)
# Returns:
#   ($ErrCode, $Mode8bit, $Flags)
#   with:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
#   $Mode8bit (boolean): TRUE if the M50530 is in 8-bit mode, FALSE if in 4-bit mode.
#   $Flags (byte): the flags as returned by the M50530's RB command
{
  my $StartTime;
  my $Flags;

  $StartTime = [gettimeofday];
  do
  {
    $Flags = RawExecuteInstr($main::READ_FLAGS, 0);  # Read the M50530 flags
    if ((!($Flags & $main::FLAG_IO8_4)) && (!($Flags & $main::FLAG_4u2_HI)))
    {
      # First quartet in 4-bit mode: get the second quartet
      $Flags = $Flags | (RawExecuteInstr($main::READ_FLAGS, 0) >> 4);
    }
    if ($Flags & $main::M50530_FLAG_BUSY_LO)
    {
      # The M50530 is busy...
      if (tv_interval($StartTime) >= $TimeOut)
      {
        # Timeout delay has elapsed
        return ($main::LCD_ERR_TIMEOUT, (($Flags & $main::FLAG_IO8_4) != 0), $Flags)  # Exit and return the TIMEOUT error code
      }
      &$WaitSub($WaitSubParms);  # Call the wait sub
    }
  } until (!($Flags & $main::M50530_FLAG_BUSY_LO));  # Wait for BUSY to get low
  return ($main::LCD_NO_ERR, (($Flags & $main::FLAG_IO8_4) != 0), $Flags);  # Indicating successful completion, return current mode and flags
}

sub ExecuteInstr
# Sends the given instruction and data to the M50530 and returns the M50530 data in case of a read instruction
# The instructions are sent taking into account the display mode (4/8-bit) and the BUSY status of the M50530
# Parameters:
#   $Instr (byte): The M50530 instruction to execute (Cf. constants / M50530 Instruction set, e.g. $main::READ_CURSOR_ADDR)
#   $Data (byte): Data to send to the M50530 in case of a write operation, disregarded in case of read operations
# Returns:
#   $RetData: the data returned by the M50530 in case of read operations, $main::LCD_NO_ERR in case of write operations
{
  my $ErrCode;
  my $Mode8Bit;
  my $RetData;  # Return value

  my ($Instr, $Data) = @_;  # Get the parameters

  ($ErrCode, $Mode8Bit) = WaitForReady;  # Wait for the M50530 to be ready
  return $ErrCode if ($ErrCode != $main::LCD_NO_ERR);  # Return the error code if not successful
  $Mode8Bit = (($Data & $main::FLAG_IO8_4) != 0) if (($Instr == 0) and (($Data & 0xC0) == 0xC0));  # Use the new mode if this is the Set Function Mode instruction
  if ($Mode8Bit)
  {
    $RetData = RawExecuteInstr($Instr, $Data);  # Single pass intruction
  }
  else
  {
    # Two pass instruction
    $RetData = RawExecuteInstr($Instr, $Data & 0xF0);  # Send / receive higher nibble
    $RetData = ($RetData | (RawExecuteInstr($Instr, $Data << 4) >> 4));  # Send / receive lower nibble
  }
  return $RetData;
}

sub SetFunctionMode
# Initializes (and clears) the LCD display
# You must call this sub before using the display
# Parameters:
#   $Mode8bit (boolean): TRUE if the display should be initialized in the 8-bit mode, FALSE for 4-bit mode
#   $Font8 (boolean): TRUE for 8 dots high font, FALSE for 12 dots high font
#   $NumLines (byte): Number of lines on the display. Valid values are 1, 2 or 4.
#   $CGRAMSize (byte): Size of the character generator RAM im words. Valid values are 0, 32, 64 or 96.
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $ModeData;

  my ($Mode8bit, $Font8, $NumLines, $CGRAMSize) = @_;  # Get the parameters

  # Set the requested mode
  $ModeData = $main::SET_FUNCTION_MODE;
  $ModeData = ($ModeData | $main::FLAG_IO8_4) if ($Mode8bit);
  $ModeData = ($ModeData | $main::FLAG_FONT8_12) if ($Font8);
  $ModeData = ($ModeData | (($NumLines << 1) & 0x0C));
  $ModeData = ($ModeData | (($CGRAMSize >> 5) & 0x03));
  return ExecuteInstr($main::MODE, $ModeData);
}

sub BoolStr
# Converts a boolean to a "TRUE" or "FALSE" string
{
  my ($b) = @_;
  if ($b)
  {
    return "TRUE";
  }
  else
  {
    return "FALSE";
  }
}

sub Bool2HL
# Converts a boolean to a "H" (high = true) or "L" (low = false) string
{
  my ($b) = @_;
  if ($b)
  {
    return "H";
  }
  else
  {
    return "L";
  }
}

sub Str2Num
# Converts a numerical string to a numeric value.
# If the string ends with "H" or "h" the result of its conversion from hexadecimal to decimal is returned
{
  my ($Data) = @_;
  if (lc(substr($Data, length($Data) - 1, 1) eq 'h'))
  {
    return hex(substr($Data, 0, length($Data) - 1));
  }
  else
  {
    return $Data;
  }
}

sub NOP
# Sends the NOP (No Operation) instruction to the display (does nothing)
{
  return ExecuteInstr($main::MODE, $main::NOP);
}

sub Home
# Sets the cursor and display addresses to 0
{
  return ExecuteInstr($main::MODE, $main::HOME);
}

sub SetBlinkFrequency
# Sets the blink frequency of the cursor or character at cursor position
# Parameters:
#   $BlinkRate (byte): 1, 2, 4 or 8. The resulting blink frequency is $BlinkRate x fOsc/4718592, where fOsc is the M50530's oscillator frequency
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($BlinkRate) = @_;  # Get the parameters

  $BlinkRate = int(log($BlinkRate)/log(2));  # Convert to 0..3

  $Data = $main::SET_BLINK_FREQ | ($BlinkRate & 0x03);
  return ExecuteInstr($main::MODE, $Data);
}

sub WriteUnderlineBit
# Sets or resets the underline bit at the cursor address and eventually increments the cursor address
# Parameters:
#   $Underline (boolean): TRUE = set the underlining, FALSE = reset the underlining
#   $CursorInc (boolean): TRUE = Increment the cursor address after setting the underline bit, FALSE = leave the cursor address unchanged
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($Underline, $CursorInc) = @_;  # Get the parameters

  $Data = $main::WRITE_UNDERLINE_BIT;
  $Data = ($Data | 0x02) if ($Underline);
  $Data = ($Data | 0x01) if ($CursorInc);
  return ExecuteInstr($main::MODE, $Data);
}

sub SetUnderlineMode
# Determines whether the underline is set, reset or left as is on write operations.
# Parameters:
#   $UnderlineChange (boolean): TRUE = on write operations, set the underline bit according to $Underline, FALSE = leave the underline bit unchanged
#   $Underline (boolean): TRUE = set the underline bit, FALSE = reset the underline bit (both only if $UnderlineChange is TRUE)
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($UnderlineChange, $Underline) = @_;  # Get the parameters

  $Data = $main::SET_UNDERLINE_MODE;
  $Data = ($Data | 0x02) if ($UnderlineChange);
  $Data = ($Data | 0x01) if ($Underline);
  return ExecuteInstr($main::MODE, $Data);
}

sub ShiftStartAddr
# Increments or decrements the cursor and/or display start address
# Parameters:
#   $ShiftCursor (boolean): TRUE = Shift the cursor address according to $CursorDec
#   $CursorDec (boolean): TRUE = Decrement cursor address, FALSE = increment cursor address (only if $ShiftCursor is TRUE)
#   $ShiftDisplay (boolean): TRUE = Shift the display address according to $DisplayDec
#   $DisplayDec (boolean): TRUE = Decrement display address, FALSE = increment display address (only if $ShiftDisplay is TRUE)
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($ShiftCursor, $CursorDec, $ShiftDisplay, $DisplayDec) = @_;  # Get the parameters

  $Data = $main::SHIFT_START_ADDR;
  $Data = ($Data | 0x08) if ($ShiftCursor);
  $Data = ($Data | 0x04) if ($CursorDec);
  $Data = ($Data | 0x02) if ($ShiftDisplay);
  $Data = ($Data | 0x01) if ($DisplayDec);
  return ExecuteInstr($main::MODE, $Data);
}

sub SetDisplayMode
# Turns on/off the display, cursor, cursor underlining, cursor blinking and char blinking at the cursor position
# Parameters:
#   $DisplayOn (boolean): TRUE = Turn on display
#   $CursorOn (boolean): TRUE = Show the cursor
#   $UnderlineOn (boolean): TRUE = Turn on underlining
#   $CursorBlinkOn (boolean): TRUE = Turns cursor blinking on
#   $CharBlinkOn (boolean): TRUE = Turns on character blinking
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($DisplayOn, $CursorOn, $UnderlineOn, $CursorBlinkOn, $CharBlinkOn) = @_;  # Get the parameters

  $Data = $main::SET_DISPLAY_MODE;
  $Data = ($Data | 0x10) if ($DisplayOn);
  $Data = ($Data | 0x08) if ($CursorOn);
  $Data = ($Data | 0x04) if ($UnderlineOn);
  $Data = ($Data | 0x02) if ($CursorBlinkOn);
  $Data = ($Data | 0x01) if ($CharBlinkOn);
  return ExecuteInstr($main::MODE, $Data);
}

sub SetEntryMode
# Determines if and how the cursor and display start addresses are modified on read and write operations
# Parameters:
#   $CursorDec (boolean): TRUE = decrement cursor address, FALSE = increment cursor address
#   $CursorOnWrite (boolean): TRUE = the cursor address is modified according to $CursorDec on write commands
#   $CursorOnRead (boolean): TRUE = the cursor address is modified according to $CursorDec on read commands
#   $DisplayDec (boolean): TRUE = decrement display address, FALSE = increment display address
#   $DisplayOnWrite (boolean): TRUE = the display address is modified according to $DisplayDec on write commands
#   $DisplayOnRead (boolean): TRUE = the display address is modified according to $DisplayDec on read commands
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
{
  my $Data;

  my ($CursorDec, $CursorOnWrite, $CursorOnRead, $DisplayDec, $DisplayOnWrite, $DisplayOnRead) = @_;  # Get the parameters

  $Data = $main::SET_ENTRY_MODE;
  $Data = ($Data | 0x20) if ($CursorDec);
  $Data = ($Data | 0x10) if ($CursorOnWrite);
  $Data = ($Data | 0x08) if ($CursorOnRead);
  $Data = ($Data | 0x04) if ($DisplayDec);
  $Data = ($Data | 0x02) if ($DisplayOnWrite);
  $Data = ($Data | 0x01) if ($DisplayOnRead);
  return ExecuteInstr($main::MODE, $Data);
}

sub ReadRAMData
# Reads data from the display's DD or CG RAM address designated by the current cursor address
# Parameters:
#   $NumBytes (int): The number of bytes to read from the display's RAM
# Returns:
#   $RAMData (string): the data read from the display's RAM
# Notes:
#   If you read more than 1 character from the display, the display must be configured to increment the cursor address
#   on a read operation, otherwise all characters are read from the current cursor address.
{
  my $I;
  my $RAMData;

  my ($NumBytes) = @_;  # Get the parameters

  # Read the data from the display's RAM
  $RAMData = '';
  for ($I=1; $I<=$NumBytes; $I++)
  {
    $RAMData = $RAMData . chr(ExecuteInstr($main::READ_RAM_DATA, 0));
  }
  return ($RAMData);
}

sub WriteRAMData
# Writes data to the display's DD or CG RAM address designated by the current cursor address
# Parameters:
#   $Data (string): The bytes to write to the display's RAM
# Returns:
#   $ErrCode (integer): 0 if successful, or an error code (cf. LCD_ERR_* constants)
# Notes:
#   The display must have been at least initialized using SetFunctionMode before calling this sub.
#   You won't see anything on the display if you haven't turned it on using SetDisplayMode.
#   If you write a string to the display, the display must be configured to increment the cursor address on a write
#   operation, otherwise all characters are written to the current cursor address.
{
  my $ErrCode;
  my $I;
  my $L;

  my ($Data) = @_;  # Get the parameters

  $L = length($Data);
  # Write the data to the display's RAM
  for ($I=1; $I<=$L; $I++)
  {
    $ErrCode = ExecuteInstr($main::WRITE_RAM_DATA, ord(substr($Data, $I - 1, 1)));
    return $ErrCode if ($ErrCode != $main::LCD_NO_ERR);  # Return the error code if not successful
  }
  return 0;
}

sub ClearHome
# Clears the display, sets the cursor and display addresses to 0
{
  return ExecuteInstr($main::MODE, $main::CLEAR_HOME);
}

sub ReadCursorAddr
#  Reads the cursor address which designates the RAM address
{
  return ExecuteInstr($main::READ_CURSOR_ADDR, 0);
}

sub WriteCursorAddr
#  Reads the cursor address which designates the RAM address
{
  my ($CursorAddr) = @_;  # Get the parameters

  return ExecuteInstr($main::WRITE_CURSOR_ADDR, $CursorAddr);
}

sub ReadDisplayStartAddr
#  Reads the display start address
{
  return ExecuteInstr($main::READ_DISPLAY_START_ADDR, 0);
}

sub WriteDisplayStartAddr
#  Writes the display start address
{
  my ($DisplayStartAddr) = @_;  # Get the parameters

  return ExecuteInstr($main::WRITE_DISPLAY_START_ADDR, $DisplayStartAddr);
}

sub TestInterface
# This sub can be used for debuging purposes or to test the interface before connecting the LCD display
{
  my $B1;
  my $B2;
  my $B3;
  my $B4;
  my $B5;
  my $B6;
  my $CGRAMSize;
  my $cmd;
  my $Data;
  my $DB4;
  my $DB5;
  my $DB6;
  my $DB7;
  my $ErrCode;
  my $EX;
  my $Flags;
  my $Font8;
  my $I;
  my $Instr;
  my $IOC1;
  my $IOC2;
  my $Mode8Bit;
  my $NumLines;
  my $RW;

  $RW = 1; # Read from LCD
  $IOC1 = 0;
  $IOC2 = 0;
  $EX = 0;
  $DB4 = 0;
  $DB5 = 0;
  $DB6 = 0;
  $DB7 = 0;
  SetInterface(\&ParallelPortInterfaceInit, \0);  # We use the parallel port interface

  do
  {
    &$M50530_RW_SET($RW);  # Set the R/W bit on the M50530
    &$M50530_IOC2_SET($IOC2);  # Set the I/OC2 bit on the M50530
    &$M50530_IOC1_SET($IOC1);  # Set the I/OC1 bit on the M50530
    &$M50530_EX_SET($EX);  # Set the EX bit on the M50530
    &$M50530_DB4_SET($DB4);  # Set the DB4 bit on the M50530
    &$M50530_DB5_SET($DB5);  # Set the DB5 bit on the M50530
    &$M50530_DB6_SET($DB6);  # Set the DB6 bit on the M50530
    &$M50530_DB7_SET($DB7);  # Set the DB7 bit on the M50530
    print "\n---Status:\n";
    print "INPUT:  DB4=", Bool2HL(&$M50530_DB4_GET), " DB5=", Bool2HL(&$M50530_DB5_GET), " DB6=", Bool2HL(&$M50530_DB6_GET), " DB7=", Bool2HL(&$M50530_DB7_GET), "\n";
    print "OUTPUT: DB4=", Bool2HL($DB4), " DB5=", Bool2HL($DB5), " DB6=", Bool2HL($DB6), " DB7=", Bool2HL($DB7), "\n";
    print "R/W=", Bool2HL($RW), " (H=Read from LCD) I/OC1=", Bool2HL($IOC1), " I/OC2=", Bool2HL($IOC2), " EX=", Bool2HL($EX), "\n";
    print "---Commands:\n";
    print "q = quit,  [Enter] = refresh status\n";
    print "Bit toggle commands: r = R/W, 1 = I/OC1, 2 = I/OC2, 4..7 = DB4..DB7, e = EX\n";
    print "---High level instructions:\n";
    print "w = WaitForReady, x = ExecuteInstr, sfm = SetFunctionMode, rrd = ReadRAMData\n";
    print "wrd = WriteRAMData, ch = ClearHome, rca = ReadCursorAddr, wca = WriteCursorAddr\n";
    print "rdsa = ReadDisplayStartAddr, wdsa = WriteDisplayStartAddr, sem = SetEntryMode\n";
    print "sdm = SetDisplayMode, ssa = ShiftStartAddr, sum = SetUnderlineMode\n";
    print "wub = WriteUnderlineBit, sbf = SetBlinkFrequency, h = Home, n = NOP\n";
    $cmd = <STDIN>;
    chop($cmd);
    if ($cmd eq '1')
    {
      $IOC1 = ! $IOC1;
    }
    elsif ($cmd eq '2')
    {
      $IOC2 = ! $IOC2;
    }
    elsif ($cmd eq '4')
    {
      $DB4 = ! $DB4;
    }
    elsif ($cmd eq '5')
    {
      $DB5 = ! $DB5;
    }
    elsif ($cmd eq '6')
    {
      $DB6 = ! $DB6;
    }
    elsif ($cmd eq '7')
    {
      $DB7 = ! $DB7;
    }
    elsif ($cmd eq 'ch')
    {
      # ClearHome
      $Data = ClearHome();
      printf("ClearHome return value = %u = %Xh\n", $Data, $Data);
    }
    elsif ($cmd eq 'e')
    {
      $EX = ! $EX;
    }
    elsif ($cmd eq 'h')
    {
      # ClearHome
      $Data = Home();
      printf("Home return value = %u = %Xh\n", $Data, $Data);
    }
    elsif ($cmd eq 'n')
    {
      # NOP
      $Data = NOP();
      printf("NOP return value = %u = %Xh\n", $Data, $Data);
    }
    elsif ($cmd eq 'r')
    {
      $RW = ! $RW;
    }
    elsif ($cmd eq 'rca')
    {
      # ClearHome
      $Data = ReadCursorAddr();
      printf("ReadCursorAddr return value = %u = %Xh\n", $Data, $Data);
    }
    elsif ($cmd eq 'rdsa')
    {
      # ReadDisplayStartAddr
      $Data = ReadDisplayStartAddr();
      printf("ReadDisplayStartAddr return value = %u = %Xh\n", $Data, $Data);
    }
    elsif ($cmd eq 'rrd')
    {
      # ReadRAMData
      print "ReadRAMData:\n";
      print "  Enter number of characters to read followed by 'h' if hex ([Enter] to abort): ";
      $I = <STDIN>;
      chop($I);
      if ($I ne '')
      {
        $Data = ReadRAMData(Str2Num($I));
        print "ReadRAMData($I) return value = '$Data'\n";
      }
      else
      {
        print "  ReadRAMData aborted.\n";
      }
    }
    elsif ($cmd eq 'sbf')
    {
      # SetBlinkFrequency
      print "SetBlinkFrequency:\n";
      print "  Enter blink rate 1, 2, 4 or 8 (or push [Enter] to abort): ";
      $Data = <STDIN>;
      chop($Data);
      if ($Data ne '')
      {
        print("  SetBlinkFrequency($Data) ");
        $Data = SetBlinkFrequency($Data);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  SetBlinkFrequency aborted.\n";
      }
    }
    elsif ($cmd eq 'sdm')
    {
      # SetDisplayMode
      print "SetDisplayMode:\n";
      print "  Enter values (using 0 or 1 for booleans) as follows:\n";
      print "  <DisplayOn>,<CursorOn>,<UnderlineOn>,<CursorBlinkOn>,<CharBlinkOn>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($B1, $B2, $B3, $B4, $B5) = split(/[,\n]/, <STDIN>);
      if (($B1 ne '') && ($B2 ne '') && ($B3 ne '') & ($B4 ne '') && ($B5 ne ''))
      {
        print("  SetDisplayMode($B1, $B2, $B3, $B4, $B5) ");
        $Data = SetDisplayMode($B1, $B2, $B3, $B4, $B5);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  SetDisplayMode aborted.\n";
      }
    }
    elsif ($cmd eq 'sem')
    {
      # SetEntryMode
      print "SetEntryMode:\n";
      print "  Enter values (using 0 or 1 for booleans) as follows:\n";
      print "  <CursorDec>,<CursorOnWrite>,<CursorOnRead>,<DisplayDec>,<DisplayOnWrite>,<DisplayOnRead>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($B1, $B2, $B3, $B4, $B5, $B6) = split(/[,\n]/, <STDIN>);
      if (($B1 ne '') && ($B2 ne '') && ($B3 ne '') & ($B4 ne '') && ($B5 ne '') & ($B6 ne ''))
      {
        print("  SetEntryMode($B1, $B2, $B3, $B4, $B5, $B6) ");
        $Data = SetEntryMode($B1, $B2, $B3, $B4, $B5, $B6);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  SetEntryMode aborted.\n";
      }
    }
    elsif ($cmd eq 'sfm')
    {
      # SetFunctionMode
      print "SetFunctionMode:\n";
      print "  Enter values (in decimal, using 0 or 1 for booleans) as follows:\n";
      print "  <Mode8Bit>,<Font8>,<NumLines>,<CGRAMSize>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($Mode8Bit, $Font8, $NumLines, $CGRAMSize) = split(/[,\n]/, <STDIN>);
      if (($Mode8Bit ne '') && ($Font8 ne '') && ($NumLines ne '') & ($CGRAMSize ne ''))
      {
        print("  SetFunctionMode($Mode8Bit, $Font8, $NumLines, $CGRAMSize) ");
        $Data = SetFunctionMode($Mode8Bit, $Font8, $NumLines, $CGRAMSize);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  SetFunctionMode aborted.\n";
      }
    }
    elsif ($cmd eq 'ssa')
    {
      # ShiftStartAddr
      print "ShiftStartAddr:\n";
      print "  Enter values (using 0 or 1 for booleans) as follows:\n";
      print "  <ShiftCursor>,<CursorDec>,<ShiftDisplay>,<DisplayDec>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($B1, $B2, $B3, $B4) = split(/[,\n]/, <STDIN>);
      if (($B1 ne '') && ($B2 ne '') && ($B3 ne '') & ($B4 ne ''))
      {
        print("  ShiftStartAddr($B1, $B2, $B3, $B4) ");
        $Data = ShiftStartAddr($B1, $B2, $B3, $B4);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  ShiftStartAddr aborted.\n";
      }
    }
    elsif ($cmd eq 'sum')
    {
      # ShiftStartAddr
      print "SetUnderlineMode:\n";
      print "  Enter values (using 0 or 1 for booleans) as follows:\n";
      print "  <UnderlineChange>,<Underline>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($B1, $B2) = split(/[,\n]/, <STDIN>);
      if (($B1 ne '') && ($B2 ne ''))
      {
        print("  SetUnderlineMode($B1, $B2) ");
        $Data = SetUnderlineMode($B1, $B2);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  SetUnderlineMode aborted.\n";
      }
    }
    elsif ($cmd eq 'w')
    {
      # WaitForReady
      ($ErrCode, $Mode8Bit, $Flags) = WaitForReady;
      print "WaitForReady results:\n";
      print '        ErrCode=', $ErrCode, "\n";
      print '       Mode8Bit=', BoolStr($Mode8Bit), "\n";
      print '   FLAG_BUSY_HI=', BoolStr($Flags & $main::FLAG_BUSY_HI), "\n";
      print '    FLAG_4u2_HI=', BoolStr($Flags & $main::FLAG_4u2_HI), "\n";
      print '     FLAG_IO8_4=', BoolStr($Flags & $main::FLAG_IO8_4), "\n";
      print '  FLAG_FONT8_12=', BoolStr($Flags & $main::FLAG_FONT8_12), "\n";
      print '   FLAG_BUSY_LO=', BoolStr($Flags & $main::FLAG_BUSY_LO), "\n";
      print '    FLAG_4u2_LO=', BoolStr($Flags & $main::FLAG_4u2_LO), "\n";
      print '       FLAG_RA1=', BoolStr($Flags & $main::FLAG_RA1), "\n";
      print '       FLAG_RA0=', BoolStr($Flags & $main::FLAG_RA0), "\n";
    }
    elsif ($cmd eq 'wca')
    {
      print "WriteCursorAddr:\n";
      print "  Enter cursor address followed by 'h' if hex ([Enter] to abort): ";
      $Data = <STDIN>;
      chop($Data);
      if ($Data ne '')
      {
        $ErrCode = WriteCursorAddr(Str2Num($Data));
        printf ("WriteCursorAddr($Data) return value = %u = %Xh\n", $ErrCode, $ErrCode);
      }
      else
      {
        print "  WriteCursorAddr aborted.\n";
      }
    }
    elsif ($cmd eq 'wdsa')
    {
      print "WriteDisplayStartAddr:\n";
      print "  Enter display address followed by 'h' if hex ([Enter] to abort): ";
      $Data = <STDIN>;
      chop($Data);
      if ($Data ne '')
      {
        $ErrCode = WriteDisplayStartAddr(Str2Num($Data));
        printf ("WriteDisplayStartAddr($Data) return value = %u = %Xh\n", $ErrCode, $ErrCode);
      }
      else
      {
        print "  WriteDisplayStartAddr aborted.\n";
      }
    }
    elsif ($cmd eq 'wrd')
    {
      # WriteRAMData
      print "WriteRAMData:\n";
      print "  Enter character or string ([Enter] to abort): ";
      $Data = <STDIN>;
      chop($Data);
      if ($Data ne '')
      {
        print('  WriteRAMData(' , $Data, ') ');
        $Data = WriteRAMData($Data);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  WriteRAMData aborted.\n";
      }
    }
    elsif ($cmd eq 'wub')
    {
      # WriteUnderlineBit
      print "WriteUnderlineBit:\n";
      print "  Enter values (using 0 or 1 for booleans) as follows:\n";
      print "  <Underline>,<CursorInc>[Enter]\n";
      print "  (or push [Enter] to abort): ";
      ($B1, $B2) = split(/[,\n]/, <STDIN>);
      if (($B1 ne '') && ($B2 ne ''))
      {
        print("  WriteUnderlineBit($B1, $B2) ");
        $Data = WriteUnderlineBit($B1, $B2);
        printf("return value = %u = %Xh\n", $Data, $Data);
      }
      else
      {
        print "  WriteUnderlineBit aborted.\n";
      }
    }
    elsif ($cmd eq 'x')
    {
      # ExecuteInstr
      print "ExecuteInstr:\n";
      print "  0=MODE, 1=READ_FLAGS, 2=WRITE_RAM_DATA, 3=READ_RAM_DATA, 4=WRITE_DISPLAY_START_ADDR\n";
      print "  5=READ_DISPLAY_START_ADDR, 6=WRITE_CURSOR_ADDR, 7=READ_CURSOR_ADDR\n";
      print "  Enter instruction (a to abort): ";
      $Instr = <STDIN>;
      chop($Instr);
      if (($Instr ne 'a') && ($Instr ne ''))
      {
        print "  Enter data followed by 'h' if hex (a to abort): ";
        $Data = <STDIN>;
        chop($Data);
        if (($Data ne 'a') && ($Data ne ''))
        {
          $Data = Str2Num($Data);  # Convert this in case it's hex
          printf("  ExecuteInstr(%Xh, %Xh) ", $Instr, $Data);
          $Data = ExecuteInstr($Instr, $Data);
          printf("return value = %u = %Xh\n", $Data, $Data);
        }
        else
        {
          print "  ExecuteInstr aborted.\n";
        }
      }
      else
      {
        print "  ExecuteInstr aborted.\n";
      }
    }
    elsif ($cmd ne 'q')
    {
      print "Unknown command '$cmd'!\n";
    }
  } until ($cmd eq "q");
  print "\nDone.\n";
}

sub Test_SII_C591003_Display
# This sub can be used to test a SII C591003 LCD display (3 lines x 24 chars + symbols, used for example in BOSCH Integral telephones)
# Writes "Hello world!" to the display's first line and shows a blinking cursor
{
  SetInterface(\&ParallelPortInterfaceInit, \0);  # We use the parallel port interface
  if (SetFunctionMode(0, 0, 4, 96) == $main::LCD_NO_ERR)  # Clear & initialize the display in 4-bit mode, 12-dots high font, 4 lines, 96 words character generator RAM
  {
    ClearHome;  # Clear display, set cursor and display start addresses to 0
    SetDisplayMode(1,0,1,0,1);  # Turn display on, show blinking block cursor
    if (WriteRAMData("Hello world!") == $main::LCD_NO_ERR)
    {
      print "Done!";
    }
    else
    {
      print "Could not write to the display's RAM!\n";
    }
  }
  else
  {
    print "Could not initialize the display!\n";
  }
}


TestInterface;
# Test_SII_C591003_Display;
